using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using Magiq.Support;
using Magiq.Support.DynamicClasses;
using MirrorMirror;

namespace Magiq.Databases
{
    public abstract class DynamicSelectCommandBuilder : CommandBuilder
    {
        private readonly IEnumerable<ColumnsAssignment> assignments;
        protected List<string> CalculatedColumns { get; private set; }
        private readonly ClassFactory classFactory = ClassFactory.Instance;
        private readonly List<string> constantColumns = new List<string>();
        private readonly List<string> constantSqlValues = new List<string>();
        private readonly List<object> constants = new List<object>();
        private readonly Dictionary<ColumnsAssignment, List<string>> queryDynamicClassMapping;
        

        protected DynamicSelectCommandBuilder(IOrmAdapter ormAdapter, IQueryable where,
                                              IEnumerable<ColumnsAssignment> assignments, bool addTypeReference)
            : base(ormAdapter, where, addTypeReference)
        {
            CalculatedColumns = new List<string>();
            queryDynamicClassMapping = new Dictionary<ColumnsAssignment, List<string>>();
            Bindings = new List<MemberBinding>();
            this.assignments = assignments;
        }

        protected string[] BaseSelectedValues { get; set; }
        protected Type DynamicClass { get; private set; }
        protected IList<MemberBinding> Bindings { get; private set; }
        protected ParameterExpression Parameter { get; private set; }
        protected string[] Columns { get; set; }

        protected override void Init(Type elementType)
        {
            base.Init(elementType);
            Parameter = Expression.Parameter(Where.ElementType, "p__");

            BuildDynamicClassForSelect();
            OnDynamicClassForSelectBuilt();
            ProcessColumnAssignments();
            ProcessDynamicSelect();
            ProcessConstantValues();

            Columns = CalculatedColumns.Union(constantColumns).ToArray();
            BaseSelectedValues = CalculatedSqlValues.Union(constantSqlValues).ToArray();
        }

        protected virtual void OnDynamicClassForSelectBuilt()
        {
        }

        private string AddDynamicProperty(ICollection<DynamicProperty> properties, Type type)
        {
            var name = "c" + properties.Count;
            properties.Add(new DynamicProperty(name, type));
            return name;
        }

        protected virtual List<DynamicProperty> CreateDynamicProperties()
        {
            var properties = new List<DynamicProperty>();

            foreach (var assignment in assignments)
            {
                var propertyTypeMapping = assignment.ColumnOnly
                                              ? null
                                              : OrmAdapter.GetMappingOf(assignment.PropertyType);

                var list = new List<string>();

                if (propertyTypeMapping != null && propertyTypeMapping.IsEntity)
                    foreach (var id in propertyTypeMapping.Ids)
                        list.Add(AddDynamicProperty(properties, id.Type));
                else
                    list.Add(AddDynamicProperty(properties, assignment.ValueType));
                
                queryDynamicClassMapping.Add(assignment, list);
            }
            return properties;
        }

        private void BuildDynamicClassForSelect()
        {
            var properties = CreateDynamicProperties();
            DynamicClass = classFactory.GetDynamicClass(properties);
        }

        private void ProcessConstantValues()
        {
            foreach (var constant in constants)
            {
                if (constant == null)
                {
                    constantSqlValues.Add("null");
                }
                else
                {
                    var p = Command.CreateParameter();

                    p.ParameterName = "@p" + Command.Parameters.Count;
                    p.Value = constant;
                    Command.Parameters.Add(p);
                    constantSqlValues.Add(p.ParameterName);
                }
            }
        }

        private void ProcessDynamicSelect()
        {
            var newExpression = Expression.MemberInit(Expression.New(DynamicClass), Bindings);

            var lambda = Expression.Lambda(newExpression, Parameter);

            var selectQuery = Where.Provider
                .CreateQuery(Expression.Call(typeof (Queryable),
                                             "Select",
                                             new[] {Where.ElementType, lambda.Body.Type},
                                             Where.Expression,
                                             Expression.Quote(lambda)));

            var command = OrmAdapter.GetSelectCommand(selectQuery);

            SetCommand(command, selectQuery);
        }

        private void ProcessColumnAssignments()
        {
            foreach (var columnAssignment in assignments)
            {
                if (columnAssignment.ValueExpression.IsConstant())
                {
                    foreach (var constant in GetConstantsFrom(columnAssignment.ValueExpression))
                        constants.Add(constant);

                    constantColumns.AddRange(columnAssignment.Columns);
                }
                else
                {
                    var dynamicProperties = queryDynamicClassMapping[columnAssignment];
                    var expressions = GetSelectExpressionsFrom(columnAssignment.ValueExpression).ToArray();
                    for (var i = 0; i < expressions.Length; i++)
                    {
                        var expression = expressions[i];
                        var body = expression.ChangeParameter(Parameter);
                        Bindings.Add(Expression.Bind(DynamicClass.GetProperty(dynamicProperties[i]), body));
                    }

                    CalculatedColumns.AddRange(columnAssignment.Columns);
                }
            }
        }

        private IEnumerable<object> GetConstantsFrom(LambdaExpression expression)
        {
            var valueMapping = OrmAdapter.GetMappingOf(expression.Body.Type);
            var parameters = new object[expression.Parameters.Count];
            var value = expression.Compile().DynamicInvoke(parameters);

            if (valueMapping.IsEntity)
            {
                foreach (var id in valueMapping.Ids)
                {
                    if (Equals(value, null))
                        yield return null;
                    else
                        yield return value.Get(id.PropertyName);
                }
            }
            else
            {
                yield return value;
            }
        }

        private IEnumerable<Expression> GetSelectExpressionsFrom(LambdaExpression expression)
        {
            var valueMapping = OrmAdapter.GetMappingOf(expression.Body.Type);

            if (valueMapping.IsEntity)
            {
                foreach (var id in valueMapping.Ids)
                {
                    var property = Expression.Property(expression.Body, id.PropertyName);
                    yield return property;
                }
            }
            else
            {
                yield return expression.Body;
            }
        }
    }
}