using System;
using System.Collections.Generic;
using System.Data;
using System.Data.Linq;
using System.Linq;
using System.Linq.Expressions;
using System.Reflection;
using Magiq.Databases;
using Magiq.Databases.Delete;
using Magiq.Databases.Insert;
using Magiq.Databases.SqlServer;
using Magiq.Databases.Update;
using Magiq.Delete;
using Magiq.Insert;
using Magiq.Providers.Sql.Insert;
using Magiq.Providers.Sql.Query;
using Magiq.Providers.Sql.Support;
using Magiq.Query;
using Magiq.Support;
using Magiq.Update;
using MirrorMirror;

namespace Magiq.Providers.Sql
{
    public class LinqToSqlAdapter : IOrmAdapter
    {
        private readonly DataContext dataContext;
        
        public IDialect Dialect { get { return SqlServerDialect.Instance; } }
        
        public IQueryable<TChild> GetQueryableFrom<T, TChild>(QueryInfo<T, TChild> queryInfo) where T : class
        {
            if(queryInfo.Compiled.IsEntitySet())
            {
                var collection = queryInfo.ToEntitySet();
                if(collection != null && collection.Get<bool>("IsDeferred"))
                    return new MagiqToSqlQueryable<TChild>( new EntitySetAdapter(collection));
            }

            return null;
        }

        public IQueryable<TSource> GetSourceQueryableFrom<TSource,TNew>(InsertSource<TNew> source)
        {
            var table = dataContext.GetTable(source.SourceCollectionItemType);
           return (IQueryable<TSource>)source.ApplyQueryTo(table);
        }

        public IEnumerable<string> GetOrderedValuesFrom(SelectStatementInformation select, IQueryable selectQuery)
        {
            return select;
        }

        public LinqToSqlAdapter(DataContext dataContext)
        {
            this.dataContext = dataContext;
        }

        #region IDatabaseAdapter Members

        public IDbCommand GetSelectCommand(IQueryable queryable)
        {
            return dataContext.GetCommand(queryable);
        }

        public ITypeMapping GetMappingOf(Type type)
        {
            return new SqlTypeMapping(dataContext.Mapping.GetMetaType(type));
        }

        public IEnumerable<DatabaseDelete<T>> BuildDatabaseDeletes<T>(IDelete<T> delete) where T : class
        {
            yield return new DatabaseDelete<T>(this)
                             {
                                 Source = delete.Source
                             };
        }

        public IEnumerable<DatabaseUpdate<T>> BuildDatabaseUpdates<T>(IUpdate<T> update) where T : class
        {
            var emulator = new MultipleTableUpdateEmulation<T>(update, this);
            return emulator.GetTableUpdates();
        }

        public IEnumerable<DatabaseInsert<TSource>> BuildDatabaseInserts<TSource, TDestination, TNew>(
            IInsert<TSource, TDestination, TNew> insert)
        {
            var newExpression = (MemberInitExpression)insert.NewExpression.Body;
            var mapping = GetMappingOf(newExpression.Type);
            var databaseInsert = new DatabaseInsert<TSource>
                                     {
                                         OrmAdapter = this,
                                         DestinationTypeMapping = mapping,
                                         Source = GetSourceQueryableFrom<TSource, TNew>(insert.Source)
                                     };

            foreach (MemberAssignment binding in newExpression.Bindings)
            {
                var parameter = binding.Expression.IsConstant()
                                    ? Expression.Parameter(newExpression.Type, "x")
                                    : binding.Expression.GetParameter();

                var valueExpression = Expression.Lambda(binding.Expression, parameter);
                var property = (PropertyInfo) binding.Member;

                var assignment = new InsertColumnsAssignment
                                     {
                                         Columns = mapping.ColumnsFor(property.Name),
                                         PropertyName = property.Name,
                                         PropertyType = property.PropertyType,
                                         ValueExpression = valueExpression,
                                         ValueType = property.PropertyType
                                     };

                databaseInsert.ColumnAssignments.Add(assignment);
            }

            //inheritance
            var destinationInheritance = new TypeInheritanceMapping(newExpression.Type, dataContext);

            if (destinationInheritance.IsInheritance)
            {
                var discriminatorProperty = destinationInheritance.DiscriminatorProperty;
                var valueExpression = Expression.Lambda(Expression.Constant(destinationInheritance.DiscriminatorCode),
                                                        Expression.Parameter(destinationInheritance.Type, "x"));

                databaseInsert.ColumnAssignments.Add(new InsertColumnsAssignment
                                                         {
                                                             Columns = mapping.ColumnsFor(discriminatorProperty.Name),
                                                             PropertyName = discriminatorProperty.Name,
                                                             PropertyType = discriminatorProperty.PropertyType,
                                                             ValueExpression = valueExpression,
                                                             ValueType = discriminatorProperty.PropertyType
                                                         });
            }

            //Collections
            EntitySetAdapter entitySetAdapter = null;

            if (insert.Destination is IEntitySetAdapterProvider)
                entitySetAdapter = ((IEntitySetAdapterProvider) insert.Destination).EntitySetAdapter;
            else if (insert.Destination.IsDeferredEntitySet())
                entitySetAdapter = new EntitySetAdapter(insert.Destination);

            if (entitySetAdapter != null)
            {
                var member = entitySetAdapter.Member;

                for (var i = 0; i < member.Association.ThisKey.Count; i++)
                {
                    var parentKeyValue = entitySetAdapter.Instance.Get(member.Association.ThisKey[i].Name);
                    var valueExpression = Expression.Lambda(Expression.Constant(parentKeyValue),
                                                            Expression.Parameter(typeof (TNew), "x"));
                    var childKeyProperty = (PropertyInfo) member.Association.OtherKey[i].Member;

                    databaseInsert.ColumnAssignments.Add(new InsertColumnsAssignment
                                                             {
                                                                 Columns = mapping.ColumnsFor(childKeyProperty.Name),
                                                                 PropertyName = childKeyProperty.Name,
                                                                 PropertyType = childKeyProperty.PropertyType,
                                                                 ValueExpression = valueExpression,
                                                                 ValueType = childKeyProperty.PropertyType
                                                             });
                }
            }

            yield return databaseInsert;
        }

        public IInsertDestination<TNew> GetInsertDestination<TSource, TDestination, TNew>(
            IInsert<TSource, TDestination, TNew> insert)
            where TDestination : class
            where TNew : class, TDestination, new()
        {
            var destination = insert.Destination;

            if (destination.IsTable())
                return new InsertTable<TNew>((ITable) destination, this);

            var entitySet = destination as IEntitySetAdapterProvider;
            if (entitySet != null)
                return new EntitySetInsertDestination<TDestination, TNew>((EntitySet<TDestination>) entitySet.EntitySetAdapter.EntitySet);

            return null;
        }
        
        public IDbConnection GetUnderlyingDatabaseConnection()
        {
            return dataContext.Connection;
        }

        #endregion
    }
}